/**
 * The MySensors Arduino library handles the wireless radio link and protocol
 * between your home built sensors/actuators and HA controller of choice.
 * The sensors forms a self healing radio network with optional repeaters. Each
 * repeater and gateway builds a routing tables in EEPROM which keeps track of the
 * network topology allowing messages to be routed to nodes.
 *
 * Created by Henrik Ekblad <henrik.ekblad@mysensors.org>
 * Copyright (C) 2013-2015 Sensnology AB
 * Full contributor list: https://github.com/mysensors/Arduino/graphs/contributors
 *
 * Documentation: http://www.mysensors.org
 * Support Forum: http://forum.mysensors.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 *******************************
 *
 * DESCRIPTION
 *
 * Controlling an electric radiator.
 * Measure temperature. Compare to the set temperature value.
 * Using some algorithm to turn radiator on or off.
 */


// Enable debug prints to serial monitor
#define MY_DEBUG

// Enable and select radio type attached
#define MY_RADIO_NRF24
//#define MY_RADIO_RFM69

#include <SPI.h>
#include <MySensors.h>
#include <DallasTemperature.h>
#include <OneWire.h>
#include <stdlib.h>
#include <string.h>

#define SKETCH_VERSION "0.13"

#define CHILD_ID_HEATER 0
#define CHILD_ID_POWER 1
#define CHILD_ID_INFO 2
#define CHILD_ID_PIR 3
#define RADIATOR_PIN 3 // Output: radiator on/off
#define TEMP_PIN 4     // Input: DS1820B temperature sensor
#define PIR_PIN 5      // Input: PIR motion sensor

#define TEMPERATURE_STR_LEN 6 // "-15.32"
#define TEMPERATURE_DECIMALS 2

typedef struct
{
    uint8_t saved;
    float setpoint;
    uint16_t radiatorMaxPower;
    float temperatureAdjustment;
    float pBand;
    uint16_t iTime;
    float dTime;
    bool pir;
} params_t;
static params_t params;

static uint8_t errors = 0;
#define ERR_NO_TEMP_SENSOR     0x01
#define ERR_FAILED_REQUEST_SP  0x02
#define ERR_FAILED_REQUEST_KWH 0x04
#define ERR_BAD_TEMP_READING   0x08

OneWire oneWire(TEMP_PIN);
DallasTemperature sensors(&oneWire);

MyMessage msgSetPoint(CHILD_ID_HEATER, V_HVAC_SETPOINT_HEAT);
static bool setpointReceived = false;

static float energy = 0.0;

void before()
{
    sensors.begin();
}

void setup()
{
    static const uint8_t MAX_RETRY = 4;
    bool succeeded = false;
    uint8_t retry = 0;
    
    randomSeed(analogRead(0));
    loadFromEeprom((char*) &params, sizeof(params));
    if (params.saved != 1)
    {
        // EEPROM empty, save default values
        params.saved = 1;
        params.setpoint = 20.0;
        params.radiatorMaxPower = 1000; // Watt, just for calculating power and energy use
        params.temperatureAdjustment = 0.0; // Degrees, added to sensor reading
        params.pBand = 1.5; // Degrees
        params.iTime = 30; // Minutes
        params.dTime = 50.0 * 1000.0; // Milliseconds
        params.pir = false;
        saveInEeprom((char*) &params, sizeof(params));
    }
#ifdef MY_DEBUG
    Serial.print("setup: setpoint=");
    Serial.println(params.setpoint);
    Serial.print("setup: radiatorMaxPower=");
    Serial.println(params.radiatorMaxPower);
    Serial.print("setup: temperatureAdjustment=");
    Serial.println(params.temperatureAdjustment);
    Serial.print("setup: pBand=");
    Serial.println(params.pBand);
    Serial.print("setup: iTime=");
    Serial.println(params.iTime);
    Serial.print("setup: dTime=");
    Serial.println(params.dTime);
    Serial.print("setup: pir=");
    Serial.println(params.pir);
#endif

    // Request setpoint for temperature from gateway
    while (!succeeded && retry++ < MAX_RETRY)
    {
        succeeded = request(CHILD_ID_HEATER, V_HVAC_SETPOINT_HEAT);
    }
    if (!succeeded)
    {
        errors |= ERR_FAILED_REQUEST_SP;
    }

    // Request energy used from gateway
    // TODO: why don't we ever get this from controller?
    succeeded = false; retry = 0;
    while (!succeeded && retry++ < MAX_RETRY)
    {
        succeeded = request(CHILD_ID_POWER, V_KWH);
    }
    if (!succeeded)
    {
        errors |= ERR_FAILED_REQUEST_KWH;
    }

    pinMode(RADIATOR_PIN, OUTPUT); // switch radiator on/off
    digitalWrite(RADIATOR_PIN, LOW);
    pinMode(PIR_PIN, INPUT);

    // requestTemperatures() will not block current thread
    sensors.setWaitForConversion(false);
    int numOneWireDevices = sensors.getDeviceCount();
#ifdef MY_DEBUG
    Serial.print("setup: Found ");
    Serial.print(numOneWireDevices);
    Serial.println(" OneWire devices");
#endif
    if (numOneWireDevices <= 0)
    {
        errors |= ERR_NO_TEMP_SENSOR;
    }
    sendParams();
}

void presentation() {
    // Register sensor to gw (they will be created as child devices)
    sendSketchInfo("MyQTherm Sensor", SKETCH_VERSION);
    present(CHILD_ID_HEATER, S_HEATER);
    present(CHILD_ID_POWER, S_POWER);
    present(CHILD_ID_INFO, S_INFO);
    present(CHILD_ID_PIR, S_MOTION);
}

// Handle incoming messages from the MySensors Gateway
void receive(const MyMessage &message) {
    if (message.type == V_HVAC_SETPOINT_HEAT)
    {
        params.setpoint = roundTemperature(message.getFloat());
        setpointReceived = true;
#ifdef MY_DEBUG
        Serial.print("receive: setpoint=");
        Serial.println(params.setpoint);
#endif
        // need to report back to controller for update
        send(msgSetPoint.set(params.setpoint, 1));
        saveInEeprom((char*) &params, sizeof(params));
    }
    else if (message.type == V_KWH)
    {
        if (message.getFloat() > 0)
        {
            energy += message.getFloat();
        }
#ifdef MY_DEBUG
        Serial.print("receive: V_KWH=");
        Serial.println(message.getFloat());
#endif
    }
    else if (message.type == V_VAR1)
    {
#ifdef MY_DEBUG
        Serial.print("receive: V_VAR1=");
        Serial.println(message.getString());
#endif
        // With serial gateway you can send data to any MySensors node
        // echo -n "message\n" > /dev/ttyUSB0
        // echo -n "1;1;1;0;24;SP=20.0\n" > /dev/ttyUSB0
        // See https://www.mysensors.org/download/serial_api_20 for API
        // node-id;child-sensor-id;command;ack;type;payload\n
        // substitute node-id and child-sensor-id
        setParameter(message.getString());
    }
#ifdef MY_DEBUG
    else
    {
        Serial.print("receive: unexpected message, type=");
        Serial.println(message.type);
    }
#endif
}

void loop() 
{
    unsigned long currentTime = millis();

    float temperature;
    if ((errors & ERR_NO_TEMP_SENSOR) == 0) {
        temperature = fetchTemperature(currentTime);
    }
    controlRadiator(currentTime, temperature);
    if (params.pir)
    {
        checkPir(currentTime);
    }
}

void setParameter(const char* buffer)
{
    char* np;
    char* name = strtok_r(buffer, "=", &np);
    char* value = strtok_r(NULL, "=", &np);
    char* confirmMsg = NULL;
    char text[MAX_PAYLOAD+1];
#ifdef MY_DEBUG
    Serial.print("setParameter: name=");
    Serial.println(name);
    Serial.print("setParameter: value=");
    Serial.println(value);
#endif
    if (value)
    {
        if (strcmp(name, "P") == 0)
        {
            // degrees C
            params.pBand = strtod(value, NULL);
#ifdef MY_DEBUG
            Serial.print("setParameter: pBand=");
            Serial.println(params.pBand);
#endif
        }
        else if (strcmp(name, "I") == 0)
        {
            // time in minutes
            params.iTime = strtoul(value, NULL, 10);
#ifdef MY_DEBUG
            Serial.print("setParameter: iTime=");
            Serial.println(params.iTime);
#endif
        }
        else if (strcmp(name, "D") == 0)
        {
            // time in milliseconds
            params.dTime = strtod(value, NULL);
#ifdef MY_DEBUG
            Serial.print("setParameter: dTime=");
            Serial.println(params.dTime);
#endif
        }
        else if (strcmp(name, "SP") == 0)
        {
            // setpoint in degrees
            params.setpoint = strtod(value, NULL);
#ifdef MY_DEBUG
            Serial.print("setParameter: setpoint=");
            Serial.println(params.setpoint);
#endif
            // need to report back to controller for update
            send(msgSetPoint.set(params.setpoint, 1));
        }
        else if (strcmp(name, "TA") == 0)
        {
            // temperature adjustment, add to current value so you
            // don't have to know the current value
            // But first check that it currently is a valid value
            if (!(params.temperatureAdjustment > -30.0 &&
                  params.temperatureAdjustment < 30.0))
            {
                params.temperatureAdjustment = 0.0;
            }
            params.temperatureAdjustment += strtod(value, NULL);
#ifdef MY_DEBUG
            Serial.print("setParameter: temperatureAdjustment=");
            Serial.println(params.temperatureAdjustment);
#endif
        }
        else if (strcmp(name, "MP") == 0)
        {
            // radiator max power in Watt
            params.radiatorMaxPower = strtoul(value, NULL, 10);
#ifdef MY_DEBUG
            Serial.print("setParameter: radiatorMaxPower=");
            Serial.println(params.radiatorMaxPower);
#endif
        }
        else if (strcmp(name, "PIR") == 0)
        {
            // motion sensor 1/0 (on/off)
            params.pir = (value[0] == '1');
#ifdef MY_DEBUG
            Serial.print("setParameter: pir=");
            Serial.println(params.pir);
#endif
        }
        else
        {
            confirmMsg = "unknown parameter";
        }

        saveInEeprom((char*) &params, sizeof(params));

        if (! confirmMsg)
        {
            sprintf(text, "%s=%s", name, value);
            confirmMsg = text;
        }
        send(MyMessage(CHILD_ID_INFO, V_TEXT).set(confirmMsg));
    }
    else
    {
        sendParams();
    }
}

enum fetchTempState_t {READY, CONVERTING, WAIT};

/*
 * Fetches temperature from sensor. Since readings fluctuates, this method makes
 * a weighted average over last reading and historical readings.
 */
float fetchTemperature(const unsigned long currentTime)
{
    // The wait need to be at least a bit longer than conversionTime
    // i.e. you cannot sample temperature faster than conversionTime
    static const uint16_t WAIT_BETWEEN_TEMP_MEASUREMENTS = 2000;
    static const uint16_t WAIT_BETWEEN_TEMP_SEND = 20000;
    static const uint16_t conversionTime =
        millisToWaitForConversion(sensors.getResolution());
    static const float alpha = 0.7; // weight for last reading
    static fetchTempState_t state = READY;
    static unsigned long nextEvent = 0;
    static unsigned long nextSend = 0;
    static float lastGoodTemp = 20.0;
    static float errorReadings = 0.0;
    static const float beta = 0.5; // weight for error reading
    float temperature = lastGoodTemp;

    if (state == READY)
    {
        sensors.requestTemperatures();
        state = CONVERTING;
        nextEvent = currentTime + conversionTime;
    }
    else if (state == CONVERTING && (long) (currentTime - nextEvent) > 0)
    {
        float tmp = sensors.getTempCByIndex(0);
        bool validTemp = (tmp != -127.00 && tmp != 85.00);
        tmp += params.temperatureAdjustment;
#ifdef MY_DEBUG
        Serial.print("fetchTemperature: temperature=");
        Serial.println(tmp);
#endif
        if (validTemp)
        {
            temperature = alpha * tmp + (1.0 - alpha) * temperature;
            lastGoodTemp = temperature;
            if ((long) (currentTime - nextSend) > 0)
            {
                send(MyMessage(CHILD_ID_HEATER, V_TEMP).set(
                         roundTemperature(temperature), 1));
                nextSend = currentTime + WAIT_BETWEEN_TEMP_SEND;
            }
        }
        else
        {
#ifdef MY_DEBUG
            Serial.println("fetchTemperature: temperature out of range");
#endif
            temperature = lastGoodTemp;
        }
        errorReadings = beta * (validTemp ? 0.0 : 1.0) + (1.0 - beta) * errorReadings;
        if (errorReadings > 0.9)
        {
            errors |= ERR_BAD_TEMP_READING;
        }
        else
        {
            errors &= ~ERR_BAD_TEMP_READING;
        }
        state = WAIT;
        nextEvent = currentTime + WAIT_BETWEEN_TEMP_MEASUREMENTS - conversionTime;
    }
    else if (state == WAIT && (long) (currentTime - nextEvent) > 0)
    {
        state = READY;
    }
    return temperature;
}

/*
 * Round a temperature value to one decimal
 */
float roundTemperature(const float temp)
{
    return static_cast<float>(static_cast<int>(temp * 10.0 + 0.5)) / 10.0;
}

// returns number of milliseconds to wait till conversion is complete (based on IC datasheet)
uint16_t millisToWaitForConversion(const uint8_t bitResolution)
{
    switch (bitResolution)
    {
    case 9:
        return 94;
    case 10:
        return 188;
    case 11:
        return 375;
    default:
        return 750;
    }
}

enum controlRadiatorState_t {STARTUP, OFF, ON};
/*
 * Turns radiator on and off during a 30s period (PERIOD).
 */
void controlRadiator(const unsigned long currentTime, const float currentTemperature)
{
    // After a power break it could be good to wait a while before
    // turning radiators on. And also to receive setpoint from Controller.
    static const unsigned long PERIOD = 30000; // 30 s period for radiator to turn on/off
    static const unsigned long WAIT_FOR_START = 5000 + random(PERIOD);
    static controlRadiatorState_t state = STARTUP;
    static unsigned long nextEvent = WAIT_FOR_START;
    static uint8_t outputPower = 0;

    if (state == STARTUP && (long) (currentTime - nextEvent) > 0)
    {
        if (!setpointReceived)
        {
            // maybe first time, not created yet
            send(msgSetPoint.set(params.setpoint, 1));
        }
        state = OFF;
    }
    else if (state == OFF && (long) (currentTime - nextEvent) > 0)
    {
        if ((errors & ERR_NO_TEMP_SENSOR) == 0)
        {
            outputPower = calcOutputPower(params.setpoint - currentTemperature, (float) PERIOD);
        }
        else
        {
            outputPower = 0;
            sendPidInfo(0, 0, 0, 0);
        }
#ifdef MY_DEBUG
        Serial.print("controlRadiator: outputPower=");
        Serial.println(outputPower);
#endif
        send(MyMessage(CHILD_ID_POWER, V_WATT).set(
                 (long) outputPower * params.radiatorMaxPower / 100));
        if (outputPower > 0)
        {
            unsigned long time = outputPower * PERIOD / 100;
            energy += (float) params.radiatorMaxPower / 1000 * time / 1000 / 3600;
            send(MyMessage(CHILD_ID_POWER, V_KWH).set(energy, 4));
            nextEvent = currentTime + time;
            state = ON;
#ifdef MY_DEBUG
            Serial.println("controlRadiator: radiator ON");
#endif
            digitalWrite(RADIATOR_PIN, HIGH);
        }
        else
        {
            nextEvent = currentTime + PERIOD;
#ifdef MY_DEBUG
            Serial.println("controlRadiator: radiator OFF");
#endif
            // The radiator could actually be on here
            // if prev outputPower was 100
            digitalWrite(RADIATOR_PIN, LOW);
        }
    }
    else if (state == ON && (long) (currentTime - nextEvent) > 0)
    {
        nextEvent = currentTime + (100 - outputPower) * PERIOD / 100;
        state = OFF;
        // Do not shut off if power is 100, since this will be unnecessary
        // if power will be > 0 next period
        if (outputPower < 100)
        {
#ifdef MY_DEBUG
            Serial.println("controlRadiator: radiator OFF");
#endif
            digitalWrite(RADIATOR_PIN, LOW);
        }
    }
}

/*
 * Calculate the needed power for the radiator in percent
 * of full power, 0-100.
 * PI regulation. P-band is 1.5 degrees. 50% at 0.75 diff.
 * I-time 30 minutes. 1% for every 36s at 0.75 diff. 50% after 30 minutes.
 * Parameter e_k: error, difference between setpoint and actual temperature.
 */
uint8_t calcOutputPower(const float e_k, const float period)
{
    const float K = 100.0 / params.pBand;
    const float I_factor = K * period / (params.iTime * 60.0 * 1000.0);
    const float D_factor = K * params.dTime / period;
    static float I_k = 0.0;
    static float e_k_1 = 0.0;

    // TODO: I_k is restricted to [-50..+50]. But maybe it is better
    // to restrict I_k further if K*e_k is at a maximum?
    // 
    I_k += I_factor * e_k;
    I_k = min(50.0, max(-50.0, I_k));
    float D_k = D_factor * (e_k - e_k_1);
    float u_k = K * e_k + I_k + D_k;
    e_k_1 = e_k;
    // use signed int here since this calculation can result in negative value
    int tmp = 50 + static_cast<int>(u_k + 0.5);
    uint8_t power = min(100, max(0, tmp));
#ifdef MY_DEBUG
    Serial.print("calcOutputPower: e_k=");
    Serial.println(e_k);
    Serial.print("calcOutputPower: I_k=");
    Serial.println(I_k);
    Serial.print("calcOutputPower: u_k=");
    Serial.println(u_k);
    Serial.print("calcOutputPower: power=");
    Serial.println(power);
#endif
    sendPidInfo(e_k, I_k, D_k, u_k);
    return power;
}

void sendPidInfo(const float e_k, const float I_k, const float D_k,
                 const float u_k)
{
    static uint8_t index = 0;
    char text[MAX_PAYLOAD+1];
    char str[11];
    float value;
    char* label;
    switch (index) {
    case 0:
        value = e_k;
        label = "e_k";
        break;
    case 1:
        value = I_k;
        label = "I_k";
        break;
    case 2:
        value = D_k;
        label = "D_k";
        break;
    case 3:
        value = u_k;
        label = "u_k";
        break;
    }
    dtostrf(value, 10, 3, str);
    sprintf(text, "%d %s=%s", errors, label, str);
    send(MyMessage(CHILD_ID_INFO, V_TEXT).set(text));
    index = ++index % 4;
}

void sendParams()
{
    char text[MAX_PAYLOAD+1];
    char str[11];

    dtostrf(params.setpoint, 10, 2, str);
    sprintf(text, "SP=%s", str);
    send(MyMessage(CHILD_ID_INFO, V_TEXT).set(text));

    sprintf(text, "MP=%d", params.radiatorMaxPower);
    send(MyMessage(CHILD_ID_INFO, V_TEXT).set(text));

    dtostrf(params.temperatureAdjustment, 10, 2, str);
    sprintf(text, "TA=%s", str);
    send(MyMessage(CHILD_ID_INFO, V_TEXT).set(text));

    dtostrf(params.pBand, 10, 2, str);
    sprintf(text, "P=%s", str);
    send(MyMessage(CHILD_ID_INFO, V_TEXT).set(text));

    sprintf(text, "I=%d", params.iTime);
    send(MyMessage(CHILD_ID_INFO, V_TEXT).set(text));

    dtostrf(params.dTime, 10, 0, str);
    sprintf(text, "D=%s", str);
    send(MyMessage(CHILD_ID_INFO, V_TEXT).set(text));

    sprintf(text, "PIR=%d", params.pir);
    send(MyMessage(CHILD_ID_INFO, V_TEXT).set(text));
}

enum pirState_t {PIR_READY, PIR_WAIT};

void checkPir(const unsigned long currentTime)
{
    static const unsigned long PERIOD = 1000; // 1 second between check
    static unsigned long nextEvent = 0;
    static pirState_t state = PIR_READY;
    static bool lastTrippedState = false;
    
    if (state == PIR_READY && (long) (currentTime - nextEvent) > 0)
    {
        // check state of pir
        bool tripped = digitalRead(PIR_PIN) == HIGH;
        if (tripped != lastTrippedState)
        {
#ifdef MY_DEBUG
            Serial.print("checkPir: tripped=");
            Serial.println(tripped);
#endif
            send(MyMessage(CHILD_ID_PIR, V_TRIPPED).set(tripped));
            lastTrippedState = tripped;
        }
        nextEvent = currentTime + PERIOD;
        state = PIR_WAIT;
    }
    if (state == PIR_WAIT && (long) (currentTime - nextEvent) > 0)
    {
        state = PIR_READY;
    }
}

void saveInEeprom(char* pVal, const uint8_t size)
{
    for (unsigned int t=0; t<size; t++)
    {
        saveState(EEPROM_LOCAL_CONFIG_ADDRESS + t, *(pVal + t));
    }
}

void loadFromEeprom(char* pVal, const uint8_t size)
{
    for (unsigned int t=0; t<size; t++)
    {
        pVal[t] = loadState(EEPROM_LOCAL_CONFIG_ADDRESS + t);
    }
}
